/*
 * Termios command line History and Editting.
 *
 * Copyright (c) 1986-2001 may safely be consumed by a BSD or GPL license.
 * Written by:   Vladimir Oleynik <vodz@usa.net>
 *
 * Used ideas:
 *      Adam Rogoyski    <rogoyski@cs.utexas.edu>
 *      Dave Cinege      <dcinege@psychosis.com>
 *      Jakub Jelinek (c) 1995
 *      Erik Andersen    <andersee@debian.org> (Majorly adjusted for busybox)
 *
 * This code is 'as is' with no warranty.
 *
 *
 */
/*
   Usage and Known bugs:
   Terminal key codes are not extensive, and more will probably
   need to be added. This version was created on Debian GNU/Linux 2.x.
   Delete, Backspace, Home, End, and the arrow keys were tested
   to work in an Xterm and console. Ctrl-A also works as Home.
   Ctrl-E also works as End.

   Small bugs (simple effect):
   - not true viewing if terminal size (x*y symbols) less
     size (prompt + editor`s line + 2 symbols)
   - not true viewing if length prompt less terminal width
 */

#include <signal.h>
#include <termio.h>

#define  MAX_HISTORY   15 /* Maximum length of the linked list for the command line history */

#define ESC     27
#define DEL     127

static struct history *his_front = NULL;/* First element in command line list */
static struct history *his_end = NULL;  /* Last element in command line list */

struct history {
	char *s;
	struct history *p;
	struct history *n;
};

static
volatile int cmdedit_termw = 80;        /* actual terminal width */
static int history_counter = 0; /* Number of commands in history list */

static int cmdedit_x;           /* real x terminal position */
static int cmdedit_y;           /* pseudoreal y terminal position */
static int cmdedit_prmt_len;    /* lenght prompt without colores string */

static int cursor;              /* required global for signal handler */
static int len;                 /* --- "" - - "" - -"- --""-- --""--- */
static const char *cmdedit_prompt;/* --- "" - - "" - -"- --""-- --""--- */

/* Link into lash to reset context to 0 on ^C and such */
extern unsigned int shell_context;


static void cmdedit_setwidth(int w, int redraw_flg);

static void win_changed(int nsig)
{
	struct winsize win = { 0, 0, 0, 0 };
	static __sighandler_t previous_SIGWINCH_handler;    /* for reset */

	/*   emulate      || signal call */
	if (nsig == -SIGWINCH || nsig == SIGWINCH) {
		ioctl(0, TIOCGWINSZ, &win);
		if (win.ws_col > 0) {
			cmdedit_setwidth(win.ws_col, nsig == SIGWINCH);
		}
	}
	/* Unix not all standart in recall signal */

	if (nsig == -SIGWINCH)          /* save previous handler   */
		previous_SIGWINCH_handler = signal(SIGWINCH, win_changed);
	else if (nsig == SIGWINCH)      /* signaled called handler */
		signal(SIGWINCH, win_changed);  /* set for next call */
	else                                            /* nsig == 0 */
		/* set previous handler    */
		signal(SIGWINCH, previous_SIGWINCH_handler);  /* reset */
}

/* special for recount position for scroll and remove terminal margin effect */
static void cmdedit_set_out_char(int next_char)
{

	int c = sedtstr[cursor];

	if (c == 0)
		c = ' ';    /* destroy end char? */
	putcc(c);
	if (++cmdedit_x >= cmdedit_termw) {
		/* terminal is scrolled down */
		cmdedit_y++;
		cmdedit_x = 0;

		if (!next_char)
			next_char = ' ';
		/* destroy "(auto)margin" */
		putcc(next_char);
		putcc('\b');
	}
	cursor++;
}

/* Move to end line. Bonus: rewrite line from cursor */
static void input_end(void)
{
	while (cursor < len)
		cmdedit_set_out_char(0);
}

/* Go to the next line */
static void goto_new_line(void)
{
	input_end();
	if (cmdedit_x)
		putcc('\n');
}


static inline void beep(void)
{
	putcc('\007');
}

/* Move back one charactor */
/* special for slow terminal */
static void input_backward(int num)
{
	if (num > cursor)
		num = cursor;
	cursor -= num;          /* new cursor (in command, not terminal) */

	if (cmdedit_x >= num) {         /* no to up line */
		cmdedit_x -= num;
		if (num < 4)
			while (num-- > 0)
				putcc('\b');

		else
			warn_d("\033[%dD", num);
	} else {
		int count_y;

		if (cmdedit_x) {
			putcc('\r');    /* back to first terminal pos. */
			num -= cmdedit_x;       /* set previous backward */
		}
		count_y = 1 + num / cmdedit_termw;
		warn_d("\033[%dA", count_y);
		cmdedit_y -= count_y;
		/*  require  forward  after  uping   */
		cmdedit_x = cmdedit_termw * count_y - num;
		warn_d("\033[%dC", cmdedit_x);  /* set term cursor   */
	}
}

static char prompt0[]=" => ";

static void put_prompt(void)
{
	putcs(cmdedit_prompt);
	putcs(prompt0);
	cmdedit_x = cmdedit_prmt_len;   /* count real x terminal position */
	cursor = 0;
}

static void parse_prompt(const char *prmt_ptr)
{
	const char *t;
	cmdedit_prompt = prmt_ptr;
	cmdedit_prmt_len = strlen(prmt_ptr)+strlen(prompt0);

	t = prmt_ptr;
	while( *t =='\n') {
	    putcc('\n');
	    t++;
	    cmdedit_prmt_len--;
	}
	cmdedit_prompt = t;
	while( (t=strchr(t, '\r')) != 0) {
	    cmdedit_prmt_len -= 2;
	    t += 2;
	}
	put_prompt();
}


/* draw promt, editor line, and clear tail */
static void redraw(int y, int back_cursor)
{
	if (y > 0)                              /* up to start y */
		warn_d("\033[%dA", y);
	cmdedit_y = 0;                          /* new quasireal y */
	putcc('\r');
	put_prompt();
	input_end();                            /* rewrite */
	warn_d("\033[J");                       /* destroy tail after cursor */
	input_backward(back_cursor);
}

/* Delete the char in front of the cursor */
static void input_delete(void)
{
	int j = cursor;

	if (j == len)
		return;

	strcpy(sedtstr + j, sedtstr + j + 1);
	len--;
	input_end();                    /* rewtite new line */
	cmdedit_set_out_char(0);        /* destroy end char */
	input_backward(cursor - j);     /* back to old pos cursor */
}

/* Delete the char in back of the cursor */
static void input_backspace(void)
{
	if (cursor > 0) {
		input_backward(1);
		input_delete();
	}
}


/* Move forward one charactor */
static void input_forward(void)
{
	if (cursor < len)
		cmdedit_set_out_char(sedtstr[cursor + 1]);
}


static void cmdedit_fflush(void)
{
	set_alarm_out();
	fflush(stdout);
	if(cuslog)
	    fflush(cuslog);
	unset_alarm_out();
}


static void cmdedit_setwidth(int w, int redraw_flg)
{
    if (w > (cmdedit_prmt_len + 2)) {

	cmdedit_termw = w;

	if (redraw_flg) {
	    /* new y for current cursor */
	    int new_y = (cursor + cmdedit_prmt_len) / w;

	    /* redraw */
	    redraw((new_y >= cmdedit_y ? new_y : cmdedit_y), len - cursor);
	    cmdedit_fflush();
	}
    }
}

static void get_previous_history(struct history **hp, struct history *p)
{
	if ((*hp)->s)
		free((*hp)->s);
	(*hp)->s = strcpym(sedtstr);
	*hp = p;
}

static inline void get_next_history(struct history **hp)
{
	get_previous_history(hp, (*hp)->n);
}

/*
 * This function is used to grab a character buffer
 * from the input file descriptor and allows you to
 * a string with full command editing (sortof like
 * a mini readline).
 *
 * The following standard commands are not implemented:
 * ESC-b -- Move back one word
 * ESC-f -- Move forward one word
 * ESC-d -- Delete back one word
 * ESC-h -- Delete forward one word
 * CTL-t -- Transpose two characters
 *
 * Furthermore, the "vi" command editing keys are not implemented.
 *
 */


static char *edtstr(char *prompt, int exist)
{

	int break_out = 0;
	char c = 0;
	struct history *hp = his_end;

	/* prepare before init handlers */
	cmdedit_y = 0;  /* quasireal y, not true work if line > xt*yt */

	detect_single_mode();

	if(exist == 0) {
	    sedtstr[0] = 0;
	    len = 0;
	} else {
	    len = strlen(sedtstr);
	}

	/* Print out the command prompt */
	parse_prompt(prompt);
	/* Now initialize things */
	win_changed(-SIGWINCH);

	while (1) {
		reput_promt_alarm(prompt, prompt0, sedtstr, 1, &cursor);
		if(exist) {
		    input_end();
		    input_backward(cursor);
		    exist = 0;
		}
		cmdedit_fflush();
		c = getkey();

		switch (c) {
		case '\n':
		case '\r':
			/* Enter */
			goto_new_line();
			break_out = len;
			while(--break_out>=0 && sedtstr[break_out]==' ')
			    sedtstr[break_out]=0;
			break_out = 1;
			break;
		case 1:
			/* Control-a -- Beginning of line */
			input_backward(cursor);
			break;
		case 2:
			/* Control-b -- Move back one character */
			input_backward(1);
			break;
		case 3:
			/* Control-c -- stop gathering input */

			/* Go to the next line */
			goto_new_line();
			sedtstr[0] = 0;
			len = 0;
			break_out = 1;
			break;
		case 4:
			/* Control-d -- Delete one character, or exit
			 * if the len=0 and no chars to delete */
			if (len == 0) {
				beep();
			} else {
				input_delete();
			}
			break;
		case 5:
			/* Control-e -- End of line */
			input_end();
			break;
		case 6:
			/* Control-f -- Move forward one character */
			input_forward();
			break;
		case '\b':
		case DEL:
			/* Control-h and DEL */
			input_backspace();
			break;
		case '\t':
			/* usupport */
			beep();
			break;
		case 14:
			/* Control-n -- Get next command in history */
			if (hp && hp->n && hp->n->s) {
				get_next_history(&hp);
				goto rewrite_line;
			} else {
				beep();
			}
			break;
		case 16:
			/* Control-p -- Get previous command from history */
			if (hp && hp->p) {
				get_previous_history(&hp, hp->p);
				goto rewrite_line;
			} else {
				beep();
			}
			break;
		case 21:
			/* Control-U -- Clear line before cursor */
			if (cursor) {
				strcpy(sedtstr, sedtstr + cursor);
				redraw(cmdedit_y, len -= cursor);
			}
			break;

		case ESC:{
			/* escape sequence follows */
			c = getkey();
			/* different vt100 emulations */
			if (c == '[' || c == 'O')
				c = getkey();
			switch (c) {
			case 'A':
			/* Up Arrow -- Get previous command from history */
				if (hp && hp->p) {
					get_previous_history(&hp, hp->p);
					goto rewrite_line;
				} else {
					beep();
				}
				break;
			case 'B':
				/* Down Arrow -- Get next command in history */
				if (hp && hp->n && hp->n->s) {
					get_next_history(&hp);
					goto rewrite_line;
				} else {
					beep();
				}
				break;

			/* Rewrite the line with the selected history item */
			  rewrite_line:
				/* change command */
				len = strlen(strcpy(sedtstr, hp->s));
				/* redraw and go to end line */
				redraw(cmdedit_y, 0);
				break;
			case 'C':
			/* Right Arrow -- Move forward one character */
				input_forward();
				break;
			case 'D':
				/* Left Arrow -- Move back one character */
				input_backward(1);
				break;
			case '3':
				/* Delete */
				input_delete();
				break;
			case '1':
			case 'H':
				/* Home (Ctrl-A) */
				input_backward(cursor);
				break;
			case '4':
			case 'F':
				/* End (Ctrl-E) */
				input_end();
				break;
			default:
				if (!(c >= '1' && c <= '9'))
					c = 0;
				beep();
			}
			if (c >= '1' && c <= '9')
				do
				    c = getkey();
				while (c != '~');
			break;
		}

		default:      /* If it's regular input, do the normal thing */
			if(len == 0 && c == ' ') {
			    beep();
			    break;
			}
			if (len >= (LEN_SEDTSTR - 2))        /* Need to leave space for enter */
				break;

			len++;

			if (cursor == (len - 1)) {      /* Append if at the end of the line */
				*(sedtstr + cursor) = c;
				*(sedtstr + cursor + 1) = 0;
				cmdedit_set_out_char(0);
			} else {                        /* Insert otherwise */
				int sc = cursor;

				memmove(sedtstr + sc + 1, sedtstr + sc, len - sc);
				*(sedtstr + sc) = c;
				sc++;
				/* rewrite from cursor */
				input_end();
				/* to prev x pos + 1 */
				input_backward(cursor - sc);
			}

			break;
		}
		if (break_out)                  /* Enter is the command terminator, no more input. */
			break;

	}

	/* Handle command history log */
	if (len) {                              /* no put empty line */

		struct history *h = his_end;
		char *ss;

		ss = strcpym(sedtstr);  /* duplicate */

		if (h == 0) {
			/* No previous history -- this memory is never freed */
			h = his_front = mallocc(sizeof(struct history));
			h->n = mallocc(sizeof(struct history));

			h->p = NULL;
			h->s = ss;
			h->n->p = h;
			h->n->n = NULL;
			h->n->s = NULL;
			his_end = h->n;
			history_counter++;
		} else {
			/* Add a new history command -- this memory is never freed */
			h->n = mallocc(sizeof(struct history));

			h->n->p = h;
			h->n->n = NULL;
			h->n->s = NULL;
			h->s = ss;
			his_end = h->n;

			/* After max history, remove the oldest command */
			if (history_counter >= MAX_HISTORY) {

				struct history *p = his_front->n;

				p->p = NULL;
				free(his_front->s);
				free(his_front);
				his_front = p;
			} else {
				history_counter++;
			}
		}
	}
	win_changed(0);
	cmdedit_fflush();
	return sedtstr;
}
