#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <sys/wait.h>

#include "../crc32.h"


#define KILL_TERM 256
#define KILL_HUP  512
#define KILL_MSK  (KILL_TERM|KILL_HUP)
#define CRIT_ERR  1024
#define VERB_MSK  (KILL_MSK|CRIT_ERR)


#define defline  "/dev/ttyS0"
#define defbaud  "115200"
#define defuser  "dionis"
#define deferlog "sersync.log"


static const char usage_msg[]="\
[-l /dev/ttySX] [-b baud] [-m] [-f] [-d dial_number]\n\
[-e tty_error] [-u user] [-s hostnumber] [-c hostnumber] [-v verbose_level]\n\
[-t repiod] [-o errlog_file]\n\
Options:\n\
\t-l tty serial line for processing. Default "defline"\n\
\t-b set speed for serial line. Default "defbaud"\n\
\t-m unset carrier modem line control for cable with 7-wire lines.\n\
\t-f disable flow control for cable with 3-wire lines.\n\
\t-d number for modem dialing. Example for pulse dialing: P8W5555555.\n\
\t-e tty console for errors messages.\n\
\t-u user name for setuid and chdir to home directory. Default "defuser"\n\
\t-s server host number. Default 1\n\
\t-c client host number. Default 0\n\
\t-t ask remote time server for timesyncing every period (in hour).\n\
     Require root privilegies (for set time).\n\
\t-o file for critical error messages. Default "deferlog"\n\
\t-v verbose level. Recommended 1. 3-debug mode. Default=0";

static void perror_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
static void error_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
static void error_msg(int flgs, const char *s, ...) __attribute__ ((format (printf, 2, 3)));
static void usage(void) __attribute__ ((noreturn));

static void dialing(char *dial);

static const char *applet_name;

static FILE *erlogs;
static char errbuff[2048];
static char prt_errbuff[2048*2048];
static int epid;

static void crit_err(const char *s, ...)
{
    if(erlogs) {
	static char ctb[] = "Wed Jun 30 21:49:08 1993\n";
	time_t t;
	int l;
	va_list p;

	time(&t);
	strcpy(ctb, ctime(&t));
	l = strlen(ctb);
	if(l > 0)
		ctb[l-1] = 0;
	fprintf(erlogs, "[%d]: %s: ", epid, ctb);
	va_start(p, s);
	vfprintf(erlogs, s, p);
	va_end(p);
	fputc('\n', erlogs);
    }
}

static void set_print_errbuf(void)
{
    unsigned char *pin = (unsigned char *)errbuff;
    char *pout = prt_errbuff;
    unsigned char c;

    while( (c = *pin++) ) {
	if(c < ' ') {
	    c += '@';
	    *pout++ = '^';
	} else if(c == 0x7f) {
	    c = '?';
	    *pout++ = '^';
	} else if(c == 0x9b) {
	    c = 'e';
	    *pout++ = '^';
	}
	*pout++ = c;
    }
    *pout = c;
}

static void perror_msg_and_die(const char *s, ...)
{
	va_list p;
	int e = errno;

	fprintf(stderr, "%s: ", applet_name);
	va_start(p, s);
	vsnprintf(errbuff, sizeof(errbuff), s, p);
	va_end(p);
	set_print_errbuf();
	fprintf(stderr, "%s: %s\n", prt_errbuff, strerror(e));
	crit_err("%s: %s", errbuff, strerror(e));
	exit(1);
}

static void error_msg_and_die(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	vsnprintf(errbuff, sizeof(errbuff), s, p);
	va_end(p);
	set_print_errbuf();
	fprintf(stderr, "%s\n", prt_errbuff);
	crit_err("%s", errbuff);
	exit(1);
}


static void usage(void)
{
  fprintf(stderr, "Usage: %s\n", usage_msg);
  error_msg_and_die("Bad call, see usage message");
}

struct speed_map {
	speed_t speed;                  /* Internal form. */
	unsigned long value;            /* Numeric value. */
	int timeout;                    /* select optimal timeout ms */
};

static const struct speed_map speeds[] = {
	{B50,     50,    100000 },
	{B75,     75,     80000 },
	{B110,    110,    60000 },
	{B134,    134,    50000 },
	{B150,    150,    40000 },
	{B200,    200,    30000 },
	{B300,    300,    20000 },
	{B600,    600,    10000 },
	{B1200,   1200,    5000 },
	{B1800,   1800,    3750 },
	{B2400,   2400,    2500 },
	{B4800,   4800,    1250 },
	{B9600,   9600,     625 },
	{B19200,  19200,    300 },
	{B38400,  38400,    200 },
#ifdef B57600
	{B57600,  57600,    150 },
#endif
#ifdef B115200
	{B115200, 115200,   100 },
#endif
#ifdef B230400
	{B230400, 230400,    50 },
#endif
#ifdef B460800
	{B460800, 460800,    50 },
#endif
};

static const int NUM_SPEEDS = (sizeof(speeds) / sizeof(struct speed_map));

static int Timeout_found;

static speed_t to_baud(unsigned long baud)
{
	int i;

	for (i = 0; i < NUM_SPEEDS; ++i)
		if (baud == speeds[i].value) {
			Timeout_found = speeds[i].timeout;
			return speeds[i].speed;
		}
	return (speed_t) - 1;
}

static struct termios oldtio;
static volatile int server_pid;
static volatile int client_pid;

static void atexit_fun(void)
{
  /* restore the old port settings */
   tcsetattr(0, TCSANOW, &oldtio);
}

/*
 * Write all of the supplied buffer out to a file.
 * This does multiple writes as necessary.
 * Returns the amount written, or -1 on an error.
 */
static void full_write(int fd, const char *buf, int len, const char *em, int vl)
{
	int cc;
	int total;
	int sav_len = len;

	total = 0;

	while (len > 0) {
		cc = write(fd, buf, len);

		if (cc < 0) {
			total = -1;
			break;
		}

		buf += cc;
		total += cc;
		len -= cc;
	}

	if(total<sav_len)
	  error_msg(vl|KILL_TERM, "Can`t send all data for %s: %m", em);
}

static int pgrp;
static int verbose;


static void error_msg(int flgs, const char *s, ...)
{
	va_list p;

	va_start(p, s);
	vsnprintf(errbuff, sizeof(errbuff), s, p);
	va_end(p);
	if((flgs & ~VERB_MSK) <= verbose) {
	  fprintf(stderr, "%s: ", applet_name);
	  if(verbose==3)
	    fprintf(stderr, "(pid=%d): ", epid);
	  set_print_errbuf();
	  fprintf(stderr, "%s\n", prt_errbuff);
	}
	if((flgs & KILL_MSK) != 0) {
	  const char *sign = (flgs & KILL_HUP) ? "SIGHUP" : "SIGTERM";

	  crit_err("%s, raising %s", errbuff, sign);
	  kill(pgrp, (flgs & KILL_HUP) ? SIGHUP : SIGTERM);
	  perror_msg_and_die("Can`t killing");
	}
	if((flgs & CRIT_ERR) != 0) {
	    crit_err("%s", errbuff);
	}
}

static void sighup_handler (int signum)
{
  char *ssig = 0;

  signal(signum, SIG_IGN);

  if(signum==SIGCHLD) {
    int status;
    int pid;

loop:
    do {
      if(server_pid < 0 && client_pid < 0) {
	pid = -1;
      } else {
	error_msg(3, "loop waitpid");
	pid = waitpid(-1, &status, WNOHANG);
	error_msg(3, "waitpid returned=%d", pid);
      }
      if(pid==server_pid && server_pid>0) {
	server_pid = -1;
	if(WIFSIGNALED(status)) {
	  error_msg(1|CRIT_ERR, "Server killed (%d)", WTERMSIG(status));
	} else {
	  error_msg(1|CRIT_ERR, "Server exited code %d", WEXITSTATUS(status));
	}
      }
      if(pid==client_pid && client_pid>0) {
	client_pid = -1;
	if(WIFSIGNALED(status)) {
	  error_msg(1|CRIT_ERR, "Client killed (%d)", WTERMSIG(status));
	} else {
	  error_msg(1|CRIT_ERR, "Client exited code %d", WEXITSTATUS(status));
	}
      }
    } while(pid > 0);
  }
  if(ssig==0) {
      ssig = signum==SIGHUP ? "Sighup signaled" : "Terminate request";
      if(signum==SIGCHLD)
	signum=SIGTERM;
    } else {
	error_msg_and_die("Exiting");
  }

  signal(SIGCHLD, SIG_IGN);

  if(server_pid > 0) {
      kill(server_pid, signum);
      error_msg(2|CRIT_ERR, "Server: %s", ssig);
  }
  if(client_pid > 0) {
      kill(client_pid, signum);
      error_msg(2|CRIT_ERR, "Client: %s", ssig);
  }
  goto loop;
}

/*
 * Read all of the supplied buffer from a file.
 * This does multiple reads as necessary.
 * Returns the amount read, or -1 on an error.
 * A short read is returned on an end of file.
 */
int full_read(int fd, char *buf, int len)
{
	int cc;
	int total;

	total = 0;

	while (len > 0) {
		cc = read(fd, buf, len);

		if (cc < 0)
			return -1;

		if (cc == 0)
			break;

		buf += cc;
		total += cc;
		len -= cc;
	}

	return total;
}

static inline int MAX(int f1, int f2)
{
  return f1>=f2? f1 : f2;
}

/********************************************************************
 *
 * setdtr - control the DTR line on the serial port.
 */

void setdtr (int tty_fd, int on)
{
    int modembits = TIOCM_DTR;

    ioctl(tty_fd, (on ? TIOCMBIS : TIOCMBIC), &modembits);
}

static void drop_priv(uid_t uid, gid_t gid)
{
  setgid(gid);

  if(geteuid()!=uid) {
      if(setuid(uid)<0)
	perror_msg_and_die("Can`t drop privilegies");
      error_msg(2, "Privilegies droped");
  }
}

/* client server start routine */
static int pipe1srv[2], pipe2srv[2];   /* stdin/stdout for server process */
static int pipe1cli[2], pipe2cli[2];   /* stdin/stdout for client process */

static void start_server_and_client(struct passwd *pwd,
					char *shostn, char *chostn)
{
  char *name;
  char *s;
  int i;

  if(pipe(pipe1srv)<0 || pipe(pipe2srv)<0)
	perror_msg_and_die("Can`t start pipes for server");
  if(pipe(pipe1cli)<0 || pipe(pipe2cli)<0)
	perror_msg_and_die("Can`t start pipes for client");

  for (s = name = pwd->pw_shell; *s != '\0';)
	if (*s++ == '/')
	  name = s;

  server_pid = fork();
  if(server_pid<0)
    perror_msg_and_die("Can`t forking()");

  if(server_pid>0) {
    close(pipe1srv[1]);  /* close W, pipe1srv[0] - read  cannel */
    close(pipe2srv[0]);  /* close R, pipe2srv[1] - write cannel */

    error_msg(3, "server_pid=%d", server_pid);
    client_pid = fork();
    if(client_pid<0) {
      kill(server_pid, SIGTERM);
      perror_msg_and_die("Can`t forking()");
    }
    if(client_pid==0) {
      /* start client */
      server_pid = 0;
      signal(SIGCHLD, SIG_IGN);
      close(pipe1cli[0]);  /* close R, pipe1cli[1] - write cannel */
      close(pipe2cli[1]);  /* close W, pipe2cli[0] - read  cannel */
      /* close server`s pipes */
      for(i=0; i<2; i++) {
	close(pipe1srv[i]);
	close(pipe2srv[i]);
      }
      close(0);
      close(1);
      if(pipe2cli[0]!=0)
	dup2(pipe2cli[0], 0); /* setup stdin */
      if(pipe1cli[1]!=1)
	dup2(pipe1cli[1], 1); /* setup stdout */

      drop_priv(pwd->pw_uid, pwd->pw_gid);
      error_msg(3, "start client %s %s -eh %s", pwd->pw_shell, name, chostn);
      execl(pwd->pw_shell, name, "-eh", chostn, (char *) 0);
      perror_msg_and_die("Can`t start client");
    }
    if(client_pid>0) {
      close(pipe1cli[1]);  /* close W, pipe1cli[0] - read  cannel */
      close(pipe2cli[0]);  /* close R, pipe2cli[1] - write cannel */
      error_msg(3, "client_pid=%d", client_pid);
    }
  }
  if(server_pid==0) {
    /* start server */
    signal(SIGCHLD, SIG_IGN);
    close(pipe1srv[0]);  /* close R, pipe1srv[1] - write cannel */
    close(pipe2srv[1]);  /* close W, pipe2srv[0] - read  cannel */
    /* close client`s pipes */
    for(i=0; i<2; i++) {
	close(pipe1cli[i]);
	close(pipe2cli[i]);
    }
    close(0);
    close(1);
    if(pipe2srv[0]!=0)
	dup2(pipe2srv[0], 0); /* setup stdin */
    if(pipe1srv[1]!=1)
	dup2(pipe1srv[1], 1); /* setup stdout */
    drop_priv(pwd->pw_uid, pwd->pw_gid);
    error_msg(3, "start server %s %s -Sh %s", pwd->pw_shell, name, shostn);
    execl(pwd->pw_shell, name, "-Sh", shostn, (char *) 0);
    perror_msg_and_die("Can`t start server");
  }
}

/* Packeting routine */
#define PKC_S 1024

static unsigned char *ppack;


static inline void charout(int c)
{
	*ppack++ = (unsigned char)c;
}


static void encode_pack(int c)
{
  if(c == '@' || c == 0xef || c == 0xf0 || c == 0 || c == 'x')
    charout('@');
  if(c == 0xef)
	c = 'P';
  else if(c == 0xf0)
	c = 'A';
  else if(c == 0)
	c = '0';
  else if(c == 'x')
	c = 'X';
  charout(c);
}

static int decode_pack(unsigned char *pdata, int i, unsigned char *np)
{
    int c = pdata[i++];

    if (c == '@') {
	c = pdata[i++];
	switch(c) {
	    case '@':
		break;
	    case 'P':
		c = 0xef;
		break;
	    case 'A':
		c = 0xf0;
		break;
	    case '0':
		c = 0;
		break;
	    case 'X':
		c = 'x';
		break;
	    default:
		return -1;
	}
    }
    *np = c;
    return i - 1;
}

static int decode_data_pack(char *pdata)
{
  int c;

  while((c = *pdata++) != 'x') {
    if (c == '@') {
	c = *pdata++;
	switch(c) {
	    case '@':
		break;
	    case 'P':
		c = 0xef;
		break;
	    case 'A':
		c = 0xf0;
		break;
	    case '0':
		c = 0;
		break;
	    case 'X':
		c = 'x';
		break;
	    default:
		return -1;
	}
    }
    charout(c);
  }
  return 1;
}


#define EN(c)   (int) ((c) + '!')

static void wordout(long int word)
{
  if (word == 0) {
    charout('z');
    charout('z');
    charout('z');
    charout('z');
    charout('z');
  } else {
    int tmp = 0;

    if(word < 0) {      /* Because some don't support unsigned long */
      tmp = 32;
      word = word - (long)(85 * 85 * 85 * 85 * 32);
    }
    if(word < 0) {
      tmp = 64;
      word = word - (long)(85 * 85 * 85 * 85 * 32);
    }
    charout(EN((word / (long)(85 * 85 * 85 * 85)) + tmp));
    word %= (long)(85 * 85 * 85 * 85);
    charout(EN(word / (85 * 85 * 85)));
    word %= (85 * 85 * 85);
    charout(EN(word / (85 * 85)));
    word %= (85 * 85);
    charout(EN(word / 85));
    word %= 85;
    charout(EN(word));
  }
}


#define times85(x)      ((((((x<<2)+x)<<2)+x)<<2)+x)
#define DE(c) ((c) - '!')

static long int word;
static long int bcount;

static int decode(int c)
{
  if (c == 'z' && bcount == 0) {
      word = 0;
  } else if ((c >= '!') && (c < ('!' + 85))) {
    if (bcount == 0) {
      word = DE(c);
      ++bcount;
    } else if (bcount < 4) {
      word = times85(word);
      word += DE(c);
      ++bcount;
    } else {
      word = times85(word) + DE(c);
      bcount = 0;
    }
  } else {
    return -1;
  }
  return 1;
}


static int fill_buf(unsigned char *outbuf, unsigned char *pbuf,
					int len, unsigned char np)
{
	int n;
	unsigned long oldcrc32;
	unsigned char *pcrc;

	ppack = pbuf;

	charout(len == 0 ? 0xef : 0xf0);   /* start indicator */
	encode_pack(np);
	pcrc = ppack;
	for(n = 0; n < 5; n++)         /* crc32 4 bytes btoa make 5 bytes */
		charout(' ');
	while(len > 0) {
		encode_pack(*outbuf++);
		len--;
	}
	charout('x');
	n = (ppack - pbuf);
	ppack = pbuf;
	oldcrc32 = INITC32();
	while(n) {
		oldcrc32 = UPDC32 (*ppack++, oldcrc32);
		n--;
	}
	oldcrc32 = ENDC32(oldcrc32);
	n = (ppack - pbuf);
	charout(0);

	ppack = pcrc;
	wordout(oldcrc32);
	return n;
}


struct SLOTS {
    unsigned char packet[PKC_S];
    unsigned char npacket;
    int psize;
    int filled;
    int sended;
};

#define PING_SLOT    0
#define PONG_SLOT    1
#define GTIME_SLOT   2
#define PTIME_SLOT   3
#define SRV_SLOT     4
#define CLI_SLOT     5
#define ACK_SLOT     6
#define MAX_SLOTS    6

static struct SLOTS slots[MAX_SLOTS+1];


static int fill1_buf(unsigned char *outbuf, int len, int slot)
{
    struct SLOTS *pslots = slots + slot;
    int psize;
    unsigned char p, c;

    p = pslots->npacket;
    /* slot &= 7; */
    slot <<= 5;
    /* jump over masked chars */
    do {
	p++;
	if(p > 31)
	    p = 0;
	c = slot | p;
    } while(c == '@' || c == 0xef || c == 0xf0 || c == 0 || c == 'x');
    pslots->npacket = p;
    psize = pslots->psize = fill_buf(outbuf, pslots->packet, len, c);
    pslots->sended = 0;
    pslots->filled = 1;
    return psize;
}

#if 0
static int noised;
static int noised1;
static int noised2;
/* add noise for experimental */
static void noise(unsigned char *b, int l)
{
    static time_t t;

    if(t == 0) {
	time(&t);
	srand(t);
    }
    if( (rand()%6) == 0) {
	noised++;
	noised1 = rand()%l;
	noised2 = rand()%l;
	b[noised1] ^= 0xa5;
	b[noised2] ^= 0xa5;
	error_msg(0, "NOISE INSERTED!");
    }
}

static void denoise(unsigned char *b)
{
    if(noised) {
	noised = 0;
	b[noised1] ^= 0xa5;
	b[noised2] ^= 0xa5;
    }
}

#else
#define noise(b, l)
#define denoise(b)
#endif


static void ack(unsigned char np, const char *msger, int slot)
{
    int psize, newp;
    unsigned char p;

    newp = slot /* & 7 */;
    newp <<= 5;
    newp |= np;
    p = newp;

    psize = fill_buf(0, slots[ACK_SLOT].packet, 0, p);
    noise(slots[ACK_SLOT].packet, psize);
    full_write(1, slots[ACK_SLOT].packet, psize, msger, 1);
#if 0
    slots[ACK_SLOT].packet[psize] = 0;
    error_msg(3, "Send ack packet=%d slot=%d =%s=", np, slot, slots[ACK_SLOT].packet);
#endif
    denoise(slots[ACK_SLOT].packet);
}



#define SERB_S 2048

static unsigned char sbuff[SERB_S];



static int pop_serin(int readed, int *pready)
{
    int ready, i;

    ready = *pready;

    if(ready > (sizeof(sbuff)-readed)) {
	i = sizeof(sbuff)-readed;
	if(i <= 0) {
	    readed = 0;
	    i = ready;
	    if(i > sizeof(sbuff))
		i = sizeof(sbuff);
	}
    } else {
	i = ready;
    }
    i = full_read(0, sbuff+readed, i);
    if(i <= 0)
	error_msg(1|KILL_TERM, "Serial line sended EOF: %m");
    error_msg(3, "%d symbols from serial readed", i);
    ready -= i;
    *pready = ready;
    readed += i;
    return readed;
}

static int tico;        /* timeout counter */
static int errorsc;     /* errors counter */

static int find_valid_serin(int new_ready, char *data, unsigned char *pnp)
{

    static int readed;
    static int stat;        /* 0 - find start, 1 - find end, 2 - detected */
    static int ready;
    int i;

    ready += new_ready;
    if(stat == 0) {
loop:                                   /* find start */
	for(i = 0; i < readed; i++) {
	    if(sbuff[i] == 0xf0 || sbuff[i] == 0xef) {
		stat++;
		break;
	    }
	}
loop1:                                  /* skip before start */
	if(i) {
	    memcpy(sbuff, sbuff + i, sizeof(sbuff) - i);
	    readed -= i;
	    error_msg(1, "broken before start %d symbols skiped", i);
	    errorsc++;
	}
	if(stat == 0) {                 /* start not found */
	    if(ready <= 0)
		return 0;
loop3:
	    readed = pop_serin(readed, &ready);
	    goto loop;
	}
    }

loop2:                              /* start found, find end */
    for(i = 1; i < readed; i++) {
	if(sbuff[i] == 'x') {
	    stat++;
	    break;
	}
	if(sbuff[i] == 0xf0 || sbuff[i] == 0xef) {
	    goto loop1;             /* another start found, goto skiping */
	}
    }
    if(stat == 1) {                 /* end not found */
	    if(ready <= 0)
		return 0;
	    if(readed >= sizeof(sbuff)) {
		stat = 0;           /* buffer haven`t end, flush full */
		readed = 0;
		error_msg(1, "broken buffer, %d symbols readed", readed);
		errorsc++;
		goto loop3;
	    }
	    readed = pop_serin(readed, &ready);
	    goto loop2;
    }
    /* end found, check packet */
    {
	unsigned char np = 0;
	unsigned char slot = 0;
	long int crc_p = 0;
	char *pdata = 0;
	int  stage = 0;
	/*
	    0     - set packet and slot number
	    1-5   - get crc from packet (5 chars)
	    6     - chk first char data
	    7     - detect data packet, second and more char for packet data
	    -1    - found error, skip to end
	*/
	error_msg(3, "packet found");
	for(i = 1; sbuff[i] != 'x'; i++) {
	    if(stage == 0) {
		stage = decode_pack(sbuff, i, &np);
		if(stage > 0) {
		    i = stage;
		    stage = 1;
		    slot = np >> 5;
		    np &= 31;
		} else {
		    error_msg(3, "decode_pack(np)<0");
		}
	    } else if(stage >= 1 && stage < 6) {
		int cur_crc_chr;

		if(stage == 1)
		    bcount = 0;
		cur_crc_chr = decode(sbuff[i]);
		if(cur_crc_chr < 0) {
		    error_msg(3, "stage = %d set to -1", stage);
		    stage = cur_crc_chr;
		} else {
		    sbuff[i] = ' ';
		    stage++;
		}
	    } else if(stage == 6) {
		if(sbuff[0] == 0xef) {
		    error_msg(3, "stage = %d (0xef) set to -1", stage);
		    stage = -1;
		} else {
		    if(bcount == 0) {
			crc_p = word;
			stage = 7;
			pdata = sbuff + i;
		    } else {
			error_msg(3, "stage=%d bcount=%ld set to -1",
					stage, bcount);
			stage = -1;
		    }
		}
	    }
	}       /* for(!='x') */
	if(stage == 6) {
	    if(sbuff[0] == 0xf0) {
		error_msg(3, "stage = %d, data not found", stage);
		stage = -1; /* data not found */
	    } else {
		    if(bcount == 0) {
			crc_p = word;
		    } else {
			error_msg(3, "stage=%d (0xef) bcount=%ld, set to -1",
					stage, bcount);
			stage = -1;
		    }
	    }
	}
	if(stage >= 6) {
	    /* calculate and compare crc */
	    unsigned long oldcrc32 = INITC32();

	    for(i = 0; sbuff[i] != 'x'; i++) {
		oldcrc32 = UPDC32 (sbuff[i], oldcrc32);
	    }
	    oldcrc32 = UPDC32 (sbuff[i++], oldcrc32); /* 'x' also */
	    oldcrc32 = ENDC32(oldcrc32);
	    if(oldcrc32 != crc_p) {
		error_msg(3, "stage = %d, bad crc, oldcrc32=%lx crc_p=%lx",
				stage, oldcrc32, crc_p);
		stage = -1;
	    } else {
		tico = 0;
	    }
	} else {
		if(stage >= 0) {
		    error_msg(3, "stage = %d, don`t have all stages", stage);
		    stage = -1;         /* don`t have all stages */
		}
		i++;                /* 'x' also */
	}
	if(stage == 7) {
	    *pnp = np;
	    ppack = data;
	    charout(slot);
	    stage = decode_data_pack(pdata);
	    charout(0);
	    if(verbose==3) {
		sbuff[i-1] = 0;
		error_msg(3, "Found data packet=%d slot=%d =%sx=", np, slot, sbuff);
	    } else
		error_msg(2, "Found data packet slot=%d.", slot);
	} else if(stage > 0) {
	    if(verbose==3) {
		sbuff[i-1] = 0;
		error_msg(3, "Found ack packet=%d slot=%d =%sx=", np, slot, sbuff);
	    } else
		error_msg(2, "Found ack packet=%d slot=%d.", np, slot);
	 }

	/* skip packet */
	memcpy(sbuff, sbuff + i, sizeof(sbuff) - i);
	readed -= i;
	stat = 0;       /* packet detected, set 0 - find next */
	if(stage < 0) {
	    /* packet have error, find next */
	    error_msg(1, "broken packed, %d symbols readed", i);
	    errorsc++;
	    goto loop;
	}
	if(stage == 6) {
	    /* ack */
	    for(i = 0; i < MAX_SLOTS; i++) {
		slots[i].sended = 1;
		if(slots[i].npacket == np && i == slot) {
		    slots[i].filled = 0;
		    break;
		}
	    }
	    if(i == MAX_SLOTS) {
		error_msg(1, "Detect strange remote packet %d, slot=%d", np, slot);
#if 0
		error_msg(KILL_TERM, "Debug. Killing");
#endif
	    }
	    goto loop;  /* find next packet for mark ack */
	}
	/* 1 (returned decode_data_pack()) - found, 0 - not found data */
	return stage;
    }
}

static void srv_cli_selected(int slot, unsigned char *buffer, int fd, int *pr)
{
    if(slots[slot].filled == 0) {
	const char *SCname = slot == SRV_SLOT ? "server" : "client";
	int nread = 0;
	int ready = *pr;

	error_msg(3, "Loop selected %s", SCname);
	buffer[0] = *SCname;
	while(ready < 255 && nread == 0) {
	    if(ioctl(fd, FIONREAD, &ready) < 0)
		error_msg(1|KILL_TERM, "%s ioctl(FIONREAD, %d): %m", SCname, ready);
	    if(ready <= 0)
		error_msg(KILL_TERM, "%s sended EOF: %m", SCname);
	    nread = ready;
	}
	if(ready > 255)
		nread = 255;
	else
		nread = ready;
	nread = read(fd, buffer+1, nread);
	if(nread <= 0)
	  error_msg(KILL_TERM, "%s sended EOF: %m", SCname);
	ready -= nread;
	*pr = ready;
	buffer[0] = (unsigned char)nread;
	fill1_buf(buffer, nread+1, slot);
	error_msg(2, "%s sended for serial line: %d bytes", SCname, nread+1);
    }
}

static void gen_ping_pong_pack(int slot)
{
    unsigned char ppbuf[8];

    ppack = ppbuf;
    charout(5);
    wordout(epid);
    fill1_buf(ppbuf, (ppack - ppbuf), slot);
    error_msg(2, slot == PING_SLOT ? "ping" : "pong");
}

static void decode_ping_pong_pack(int slot, unsigned char *b, int np)
{
    static int remote_pid;
    const char *ppmsg = slot == PING_SLOT ? "ping" : "pong";
    int ic;

    bcount = 0;
    for(ic = 0; ic < 5; ic++) {
	if(decode(b[ic]) < 0) {
	    error_msg(1|KILL_TERM, "Readed %s with bad pid", ppmsg);
	}
    }
    if(b[ic] || bcount)
	error_msg(1|KILL_TERM, "Readed %s with bad pid", ppmsg);
    if(word == 0)
	error_msg(1|KILL_TERM, "Readed %s with pid=0", ppmsg);
    if(remote_pid == 0)
	remote_pid = word;
    else if(remote_pid != word)
	    error_msg(KILL_TERM, "Detect change pid from %s", ppmsg);
    ack(np, (slot == PING_SLOT ? "ack ping" : "ack pong"), slot);
}


int main(int argc, char **argv)
{
		    /* options */
  char   *line = 0;
  char   *user = 0;
  char   *eline = 0;
  char   *ofile = 0;
  int    rtscts = 1;
  unsigned int ms=1;              /* modem carrier state */
  char   *dial = 0;
  char   *shostn = 0;
  char   *chostn = 0;
  long   flg_timesync = 0;
		    /* internal options values */
  speed_t speed = (speed_t)-1;
  time_t timesync;
  time_t start_time = 0;
  struct passwd *pwd;

  struct termios newtio;        /* for set raw tty mode */

		    /* variables for select() */
  fd_set readfs;              /* file descriptor set */
  struct timeval Timeout;
  int    maxfd = 1;           /* maximum file desciptor used */

  unsigned char buffer[SERB_S];  /* serial input buffer or
				    cannel + size + 254 data + '\0' */
		    /* symbols ready for read() */
  int srv_ready = 0;
  int cli_ready = 0;
  unsigned char srv_pn_prev = 0;
  unsigned char cli_pn_prev = 0;

  unsigned char np;             /* remote packet number for ack */
  int res;                      /* termorary variable result state */
  char   *s;                    /* other temporary vars */

  epid = getpid();
  /* set programm name without path */
  for ((const char *)s = applet_name = argv[0]; *s != '\0';)
    if (*s++ == '/')
      applet_name = s;
  if (*applet_name == '-' && *(applet_name + 1) != 0) /* -shell */
	applet_name++;

  while ((res=getopt(argc, argv, "l:b:u:e:mfd:v:s:c:t:o:")) != -1) {
	switch(res) {
		case 'l':
			line = optarg;
			break;
		case 'b':
			res = strtol(optarg, &s, 10);
			if(s != optarg && *s == 0)  /* good num */
			  speed = to_baud(res);
			if(speed == (speed_t)-1)
			  error_msg_and_die("Bad speed option");
			break;
		case 'u':
			user = optarg;
			break;
		case 'e':
			eline = optarg;
			break;
		case 'o':
			ofile = optarg;
			break;
		case 'm':
			ms = 0;
			break;
		case 'f':
			rtscts = 0;
			break;
		case 'v':
			verbose = strtol(optarg, &s, 10);
			if( ! (s != optarg && *s == 0 &&
					verbose >= 0 && verbose <= 3) )
			    error_msg_and_die("Bad verbose option");
			break;
		case 'd':
			dial = optarg;
			break;
		case 's':
			shostn = optarg;
			res = strtol(optarg, &s, 10);
			if( ! (s != optarg && *s == 0 &&
						res >= 0 && res <= 255) )
			    error_msg_and_die("Bad server host number");
			break;
		case 'c':
			chostn = optarg;
			res = strtol(optarg, &s, 10);
			if( ! (s != optarg && *s == 0 &&
					res >= 0 && res <= 255) )
			    error_msg_and_die("Bad client host number");
			break;
		case 't':
			flg_timesync = strtol(optarg, &s, 10);
			if(s!=optarg && *s == 0) { /* good num */
			  if(flg_timesync<0 || flg_timesync>24)
			    error_msg_and_die("Bad timesync period (1-24)");
			}
			flg_timesync *= 60l*60l;    /* to sec */
			time(&start_time);
			start_time -= flg_timesync;
			break;
		default:
			usage ();
	}
  }

  if(optind != argc)
      usage ();
  if(dial != 0 && ms == 0) {
    error_msg(0, "Daling and unset modem lines contradictioning");
    usage ();
  }
  if(eline) {
    if(freopen(eline, "a", stderr)==NULL)
	perror_msg_and_die("Can`t open %s for writen errors messages", eline);
  }

  (void)ioctl(2, TIOCSCTTY, 0);
  setvbuf(stderr, NULL, _IONBF, BUFSIZ);

  if(line==0) {
      line = defline;
      error_msg(1, "Using %s line", line);
  }
  if(speed == (speed_t)-1) {
    res = strtol(defbaud, &s, 10);
    if(s!=optarg && *s == 0)  /* good num */
      speed = to_baud(res);
    if(speed == (speed_t)-1)
      error_msg_and_die("Require speed option");
    error_msg(1, "Using %s speed", defbaud);
  }
  if(user==0) {
    user = defuser;
    error_msg(1, "Using user as %s", user);
  }
  if(shostn==0) {
    shostn = "1";
    error_msg(1, "Using server host number as %s", shostn);
  }
  if(chostn==0) {
    chostn = "0";
    error_msg(1, "Using client host number as %s", chostn);
  }
  if(strcmp(shostn, chostn)==0) {
    error_msg_and_die("Can`t start with server=client host number");
  }
  pwd = getpwnam (user);
  if(pwd==NULL) {
    int uid = strtol(user, &s, 10);

    if(s != user && *s == 0)  /* good num */
      pwd = getpwuid (uid);
    if(pwd==NULL)
      error_msg_and_die("Can`t found user informations about %s", user);
  }
  if(chdir (pwd->pw_dir)!=0)
      perror_msg_and_die("Can`t chdir to user home directory: %s", pwd->pw_dir);
  if(pwd->pw_uid <= 0)
      perror_msg_and_die("Bad uid for running this program: %d", pwd->pw_uid);

  if(ofile==0) {
      ofile = deferlog;
      error_msg(1, "Using `%s' log file", ofile);
      erlogs = fopen(ofile, "a");
      if(erlogs == NULL) {
	perror_msg_and_die("Can`t open %s for writen log messages", ofile);
      }
      chown(ofile, pwd->pw_uid, pwd->pw_gid);
  }
  signal(SIGHUP,  sighup_handler);
  signal(SIGTERM, sighup_handler);
  signal(SIGCHLD, sighup_handler);

  (void)ioctl(0, TIOCSCTTY, 0);
  close(0);
  close(1);
  setsid();
  res = open(line, O_RDWR | O_NONBLOCK | O_NOCTTY);
  if (res != 0)
      perror_msg_and_die("Can`t open serial line %s", line);
  if(!isatty(2))
	(void)ioctl(0, TIOCSCTTY, 1);
   else
	(void)ioctl(2, TIOCSCTTY, 1);
  dup(0);
  setpgrp();
  pgrp = getpgrp();
  if(pgrp < 0)
	perror_msg_and_die("getpgrp()");
  if(pgrp != epid)
	error_msg_and_die("setpgrp()");

  tcgetattr(0, &oldtio); /* save current port settings */

  bzero(&newtio, sizeof(newtio));
  /*
    CRTSCTS : output hardware flow control (only used if the cable has
	      all necessary lines. See sect. 7 of Serial-HOWTO)
    CS8     : 8n1 (8bit,no parity, 1 stopbit)
    CLOCAL  : local connection, no modem contol
    CREAD   : enable receiving characters
  */
  /* initialize and dialing processes in local mode */
  newtio.c_cflag = speed | CS8 | CREAD | CLOCAL;
  if(rtscts)
      newtio.c_cflag |= CRTSCTS;
  newtio.c_cflag |= HUPCL;
  /*
    make device raw (no other input processing)
  */
  newtio.c_iflag = 0;
  /* newtio.c_oflag = 0; - Raw output. */
  /* newtio.c_lflag = 0; - Set input mode (non-canonical, no echo,...) */
  /* newtio.c_cc[VTIME] = 0;   inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 character arrives */

  tcflush(0, TCIFLUSH);

  atexit(atexit_fun);     /* restore the old port settings in exit process */
  tcsetattr(0, TCSANOW, &newtio);

  error_msg(2, "Serial setted");

  if(ms) {
    if(dial)
	dialing(dial);
    newtio.c_cflag &= ~CLOCAL;
    tcsetattr(0, TCSAFLUSH, &newtio);
  } /* end modem lines signalling */

  /* go to blocking input even in local mode */
  fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NONBLOCK);
  error_msg(3, "Serial setted ~O_NONBLOCK");

  if(flg_timesync <= 0) {
    /* Drop setuid privilegies, uid=0 require for settime only */
    drop_priv(pwd->pw_uid, pwd->pw_gid);
  }
				    /* loop for input */
  tico = 1;     /* initialize timeout ping/pong counter */
  gen_ping_pong_pack(PING_SLOT);
  FD_ZERO(&readfs);                 /* clear setings */
  while (1) {
    error_msg(3, "Loop");
    res = find_valid_serin(0, buffer, &np);
    if(res == 0) {
	/* check for first send or if timeout resend */
	for(res = 0; res < MAX_SLOTS; res++)
	    if(slots[res].filled) {
		    if(slots[res].sended == 0) {
			noise(slots[res].packet, slots[res].psize);
			full_write(1, slots[res].packet, slots[res].psize,
						    "serial line", 0);
			denoise(slots[res].packet);
			error_msg(3, "Sended packet %d slot %d",
				slots[res].npacket, res);
			if(res == PING_SLOT)
			    slots[res].sended = MAX_SLOTS;
			else
			    slots[res].sended = ms ? MAX_SLOTS : 1;
		    } else {
			error_msg(3, "Pending %d slot", res);
			slots[res].sended--;
		    }
	    }

	FD_SET(0, &readfs);             /* set testing for serial line */
	if(maxfd > 1) {
	    if(slots[SRV_SLOT].filled == 0 && srv_ready <= 0) {
		/* set testing for pipe server */
		    FD_SET(pipe1srv[0], &readfs);
	    } else
		FD_CLR(pipe1srv[0], &readfs);
	    if(slots[CLI_SLOT].filled == 0 && cli_ready <= 0) {
		/* set testing for pipe client */
		    FD_SET(pipe1cli[0], &readfs);
	    } else
		FD_CLR(pipe1cli[0], &readfs);
	}
	/* set timeout value within input loop */
	Timeout.tv_usec = 0;            /* microseconds */
	if(ms) {
	    Timeout.tv_sec  = 30/MAX_SLOTS; /* 30 seconds */
	} else {
	    if(maxfd > 1) {
		Timeout.tv_sec = Timeout_found/1000;
		Timeout.tv_usec  = Timeout_found * 1000 -
					Timeout.tv_sec * 1000000;
	    } else {
		Timeout.tv_sec = 1; /* before first pong */
	    }
	}

	res = select(maxfd, &readfs, NULL, NULL, &Timeout);
	if(res < 0)
	    error_msg(KILL_HUP, "select()=%d: %m", res);

	if(ms) {
	    ioctl(0, TIOCMGET, &ms);
	    if((ms & TIOCM_CAR)==0)
	    error_msg(KILL_HUP, "Carrier lost");
	}
	error_msg(3, "Loop selected %d fd", res);
	if(res == 0) {
	    static int prev_errrosc;

	    error_msg(3, "%d-th timeout occurred", tico + 1);
	    if(errorsc != prev_errrosc) {
		prev_errrosc = errorsc;
		error_msg(1, "ERRORS=%d", errorsc);
	    }
	    /* number of file descriptors with input = 0, timeout occurred. */
	    if(tico > (MAX_SLOTS * 2))      /* always timeout */
		error_msg(KILL_HUP, "Packet timeout");

	    for(res = 0; res < MAX_SLOTS; res++)
		if(slots[res].filled)
		    break;
	    if(res == MAX_SLOTS) {
		gen_ping_pong_pack(PING_SLOT);
		slots[PING_SLOT].sended = MAX_SLOTS - 1;
	    }
	    tico++;
	    continue;
	}
	res = 0;
    }
    if (res > 0 || FD_ISSET(0, &readfs)) {
	if(res == 0) {
	    /* input from serial line available */
	    error_msg(3, "Loop selected serial");
	    if(ioctl(0, FIONREAD, &res) < 0)
		error_msg(1|KILL_TERM, "serial ioctl(FIONREAD, %d): %m", res);
	    if(res <= 0)
		error_msg(1|KILL_TERM, "Serial line sended EOF: %m");
	    res = find_valid_serin(res, buffer, &np);
	}
	if(res > 0) {
	    /* found data packet */
	    int c0 = buffer[0];
	    char *b = buffer+2;
	    int l = (int)buffer[1];

	    if(c0 >= MAX_SLOTS)
		error_msg(1|KILL_TERM, "Readed %02x, not valid slot", c0);
	    if(l == 0)
		error_msg(1|KILL_TERM, "Readed packet with zero length");
	    error_msg(3, "Found data packet %c %d", c0, l);
	    if(strlen(b) != l)
		error_msg(1|KILL_TERM, "Readed invalid packet length");
	    if(c0 == PING_SLOT) {
		decode_ping_pong_pack(PING_SLOT, b, np);
		if(slots[PONG_SLOT].filled == 0)
		    gen_ping_pong_pack(PONG_SLOT);
		if(slots[PING_SLOT].sended)
		    slots[PING_SLOT].sended--;  /* up */

	    } else if(c0 == GTIME_SLOT && b[0] == 't' && b[1] == '\0') {
		ack(np, "ack get time", GTIME_SLOT);
		sprintf(buffer, " %lx", time(0));
		res = strlen(buffer);
		buffer[0] = (char)(res-1);
		fill1_buf(buffer, res, PTIME_SLOT);    /* reset to fresh */
		error_msg(1, "Put time for sync");

	    } else if(c0 == PONG_SLOT) {
		decode_ping_pong_pack(PONG_SLOT, b, np);
		if(maxfd == 1) {
		    start_server_and_client(pwd, shostn, chostn);
		    /* maximum bit entry (fd) to test readers */
		    maxfd = MAX (0, pipe1srv[0]);
		    maxfd = MAX (maxfd, pipe1cli[0]);
		    maxfd++;
		}
		if(flg_timesync) {
		    time(&timesync);
		    if((start_time+flg_timesync) <= timesync) {
			start_time = timesync;
			if(slots[GTIME_SLOT].filled == 0) {
			    fill1_buf("\001t", 2, GTIME_SLOT);
			    error_msg(1, "Get time for sync");
			} else {
			    slots[GTIME_SLOT].sended = 0;
			}
		    }
		}

	    } else if(c0 == CLI_SLOT && maxfd > 1) {
		ack(np, "ack client", CLI_SLOT);
		if(srv_pn_prev != np) {         /* tst duplicate packet */
			full_write(pipe2srv[1], b, l, "server", 1);
			srv_pn_prev = np;
		} else {
			error_msg(2, "DUP(client) packet");
		}

	    } else if(c0 == SRV_SLOT && maxfd > 1) {
		ack(np, "ack server", SRV_SLOT);
		if(cli_pn_prev != np) {         /* tst duplicate packet */
			full_write(pipe2cli[1], b, l, "client", 1);
			cli_pn_prev = np;
		} else {
			error_msg(2, "DUP(server) packet");
		}

	    } else if(c0 == PTIME_SLOT) {
		ack(np, "ack put time", PTIME_SLOT);
		error_msg(1, "Sync time");
		timesync = strtol(b, &s, 16);
		if(flg_timesync>0) { /* block set time if not specified */
		    if(s == b || *s != '\0')
			error_msg(KILL_TERM, "Invalid time for sync");
		    else if(stime(&timesync))
			error_msg(1, "Time sync failed: %m");
		}

	    } else {
		if((c0 == SRV_SLOT || c0 == CLI_SLOT) && maxfd == 1)
		    error_msg(2, "packet %c, but servers don`t started", c0);
		else
		    error_msg(1, "Invalid packet =%c %d %s=", c0, l, b);
	    }

	}
    }
    if (maxfd > 1 && (srv_ready || FD_ISSET(pipe1srv[0], &readfs))) {
	/* input from server available */
	srv_cli_selected(SRV_SLOT, buffer, pipe1srv[0], &srv_ready);
    }
    if (maxfd > 1 && (cli_ready || FD_ISSET(pipe1cli[0], &readfs))) {
	/* input from client available */
	srv_cli_selected(CLI_SLOT, buffer, pipe1cli[0], &cli_ready);
    }
  }
}

static void dialing(char *dial)
{
    int stat;
    int nread;
    char buffer[512];
    fd_set readfs;              /* file descriptor set */
    int res;
    struct timeval Timeout;

    /* pulsing DTR for reset modem */
    setdtr(0, 0);
    sleep(1);
    setdtr(0, 1);

    if(dial) {
      full_write(1, "atz\r\n", 5, "dialup reset modem", 0);
      stat = 0;
      nread = 0;
      FD_ZERO(&readfs);              /* clear setings */
      while(stat<2) {
	FD_SET(0, &readfs);            /* set testing for serial line */
	/* set timeout value within input loop */
	Timeout.tv_sec  = 3;    /* 3 seconds */
	Timeout.tv_usec = 0;    /* milliseconds */
	res = select(1, &readfs, NULL, NULL, &Timeout);
	if(res==0 || nread >= sizeof(buffer) || read(0, buffer+nread, 1) <= 0) {
	  if(stat) /* may be not get Ok\r\n */
	    break;
	  error_msg_and_die("Initialize timeout");
	}
	if(buffer[nread]=='\n') {
	  /* atz\r\n and Ok\r\n */
	  buffer[nread]=0;
	  error_msg(1, "%s", buffer);
	  nread=0;
	  stat++;
	} else {
	  nread++;
	}
      }

      full_write(1, "atd", 3, "dial", 0);
      full_write(1, dial, strlen(dial), "dial", 0);
      full_write(1, "\r\n", 2, "dial", 0);
      nread = 0;
      stat = 0;
      FD_ZERO(&readfs);              /* clear setings */
      while(1) {
	FD_SET(0, &readfs);            /* set testing for serial line */
	/* set timeout value within input loop */
	Timeout.tv_sec  = stat==0 ? 3 : 30;    /* 3-atz 30-CONNECT seconds */
	Timeout.tv_usec = 0;    /* milliseconds */
	res = select(1, &readfs, NULL, NULL, &Timeout);
	if(res==0 || nread >= sizeof(buffer) || read(0, buffer+nread, 1) <= 0)
	  error_msg_and_die("Dial timeout");
	if(buffer[nread]=='\n') {
	  /* atdXXXXXXX\r\n and CONNECT/BUSY/OTHER\r\n */
	  buffer[nread]=0;
	  error_msg(1, "%s", buffer);
	  nread=0;
	  if(stat) {
	    if(strncmp("CONNECT", buffer, 7)==0)
	      break;
	    if(strncmp("BUSY", buffer, 4)==0 ||
		  strncmp("NO DIAL", buffer, 7)==0 ||
		   strncmp("NO CAR", buffer, 6)==0 ||
		    strncmp("ERROR", buffer, 5)==0) {

			error_msg_and_die("Can`t CONNECT");
	    }
	  }
	  stat++;
	} else {
	  nread++;
	}
      }
    } /* end dial */

    for(stat=0; stat<60; stat++) {
      int cd;

      sleep(1);
      ioctl(0, TIOCMGET, &cd);
      if((cd & TIOCM_CAR))    /* carrier detect */
	break;
      error_msg(3, "\ncarrier detect not");
    }
    if(stat == 60) {   /* carrier detect not */
      error_msg_and_die("CONNECT timeout");
    }

    error_msg(1, "Serial detected CONNECT");
}
