/*      -    . */

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>

#ifdef __CYGWIN__
#include <windows.h>
#endif

#ifndef _POSIX_VDISABLE
#define _POSIX_VDISABLE '\0'
#endif

#define DATABUFSIZE 128
#define IACBUFSIZE  128

static struct termios termios_def;
static struct termios termios_raw;

//#define DEBUG

#ifndef DEBUG
#define DEB_TRACE(TRACE)
#else
#define DEB_TRACE(TRACE) TRACE
#endif

#include <stdarg.h>

void *xmalloc(size_t size);
void *xrealloc(void *old, size_t size);

static char *make_message (const char *fmt, va_list p)
{
	/* Guess we need no more than 100 chars of space. */
	int size = 100;
	char *buffer = (char *) xmalloc (size);
	while (1) {
		/* Try to print in the allocated space. */
		int nchars = vsnprintf (buffer, size, fmt, p);
		/* If that worked, return the string. */
		DEB_TRACE(fprintf(stderr, "make_message=%d\n", nchars);)
		if (nchars >= 0 && nchars<size)
			return buffer;
		/* Else try again with twice as much space. */
		size *= 2;
		buffer = (char *) xrealloc (buffer, size);
	}
}

#ifdef __CYGWIN__
static const unsigned char koi8_win[] = {
/* 128  129  130  131  132  133  134  135 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 136  137  138  139  140  141  142  143 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 144  145  146  147  148  149  150  151 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 152  153  154  155  156  157  158  159 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 160  161  162  163  164  165  166  167 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 168  169  170  171  172  173  174  175 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 176  177  178  179  180  181  182  183 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 184  185  186  187  188  189  190  191 */
     0,   0,   0,   0,   0,   0,   0,   0,
/* 192  193  194  195  196  197  198  199 */
   '', '', '', '', '', '', '', '',
/* 200  201  202  203  204  205  206  207 */
   '', '', '', '', '', '', '', '',
/* 208  209  210  211  212  213  214  215 */
   '', '', '', '', '', '', '', '',
/* 216  217  218  219  220  221  222  223 */
   '', '', '', '', '', '', '', '',
/* 224  225  226  227  228  229  230  231 */
   '', '', '', '', '', '', '', '',
/* 232  233  234  235  236  237  238  239 */
   '', '', '', '', '', '', '', '',
/* 240  241  242  243  244  245  246  247 */
   '', '', '', '', '', '', '', '',
/* 248  249  250  251  252  253  254  255 */
   '', '', '', '', '', '', '', '',
};

static void win_koi8(unsigned char *s)
{
	while(*s) {
		if(*s >= 0300) {
			int i;
			for(i=0; i<64; i++)
				if(*s == koi8_win[i+64]) {
					*s = 0300+i;
					break;
				}
		}
		s++;
	}
}

static void msg(FILE *out, const char *s, ...)
{
	va_list p;
	char *m;

	va_start(p, s);
	m = make_message(s, p);
	va_end(p);
	s = m;
	while(*s) {
		int c = *((unsigned char *)s);
		if(c>=128)
			c = koi8_win[c-128];
		putc(c, out);
		s++;
	}
	free(m);
}
#else

#define msg fprintf
#define win_koi8(s)

#endif

static int term_inited;

static void doexit(int ev)
{
	if(term_inited)
		tcsetattr(0, TCSADRAIN, &termios_def);
	exit(ev);
}

void error_msg_and_die(const char *s, ...)
{
	va_list p;
	char *m;

	va_start(p, s);
	m = make_message(s, p);
	va_end(p);
	s = m;
	msg(stderr, "%s", s);
	free(m);
	doexit(1);
}

static char buf[DATABUFSIZE];

static int netfd; /* console fd:s are 0 and 1 (and 2) */

static int remote_connect(struct in_addr addr, int port)
{
	struct sockaddr_in s_addr;
	int s;
	int one = 1;
	int setfl;
	fd_set writefds;
	struct timeval tv;

	s = socket(AF_INET, SOCK_STREAM, 0);

	setfl = fcntl(s, F_GETFL, 0);
	setfl |= O_NONBLOCK;
	fcntl(s, F_SETFL, setfl);

	memset(&s_addr, 0, sizeof s_addr);
	s_addr.sin_family = AF_INET;
	s_addr.sin_port = htons(port);

	s_addr.sin_addr = addr;

	setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &one, sizeof one);

	if (connect(s, (struct sockaddr *)&s_addr, sizeof s_addr) < 0) {
		if ((errno != EWOULDBLOCK) && (errno != EINPROGRESS))
			error_msg_and_die("Unable to connect to remote host: %s\n", strerror(errno));

		FD_ZERO(&writefds);
		FD_SET(s, &writefds);
		tv.tv_sec = 20;
		tv.tv_usec = 0;
		if(select(s+1, NULL, &writefds, NULL, &tv)<=0 || FD_ISSET(s, &writefds)==0)
			error_msg_and_die("Unable to connect to remote host: timeout\n");
	}
	setfl = fcntl(s, F_GETFL, 0);
	setfl &= ~O_NONBLOCK;
	fcntl(s, F_SETFL, setfl);

	return s;
}



static FILE *fscr;

static const char *rez_words[]= {
	"wait_or", "wait", "send", "if", "exit", "or", ":", "echo",
	"updir", "downdir", "host", "port", 0
};

static const char *rez_words_if[]= {
	"error", "success",  "wait_or", 0
};

enum CMDS {
	WAIT_OR_CMD=0, WAIT_CMD, SEND_CMD, IF_CMD, EXIT_CMD, OR_CMD,
	LABEL_SET, ECHO_CMD, UP_DIR_CMD, DOWN_DIR_CMD, HOST_CMD, PORT_CMD
};

struct SCR_S {
	enum CMDS cmd;
	int timeout_code;       /* timeout or code for exit or wait_or_num */
	char *word;             /* wait_string or label_name */
};

static struct SCR_S *scr_s;
static int n_scr_s;
static struct SCR_S *cur_scr;
static int error_wait_code = 0;
static int wait_flg;

static char **wait_words;
static int nwait_words;
static int *n_found_chars;

#ifdef __CYGWIN__
static char *download_dir = "C:/";
static char *upload_dir = "e:/usr/";

#else
static char *download_dir = "./";
static char *upload_dir = "./";
#endif
static int up_d_alloc, down_d_alloc;

static char *h = "192.168.10.9";
static char *porta = "2323";
static int h_alloc, p_alloc;


static void err_proto(void)
{
	error_msg_and_die("Client:  !\n");
}

void *xmalloc(size_t size)
{
	void *ptr = malloc(size);

	if (!ptr)
		error_msg_and_die("Memory exhausted\n");
	return ptr;
}

void *xrealloc(void *old, size_t size)
{
	void *ptr;

	ptr = realloc(old, size);
	if (!ptr)
		error_msg_and_die("Memory exhausted\n");
	return ptr;
}


extern char * xstrdup (const char *s) {
	char *t;

	if (s == NULL)
		return NULL;

	t = strdup (s);

	if (t == NULL)
		error_msg_and_die("Memory exhausted\n");

	return t;
}


static char *strscript = "scr.scr";
static int s_allocated;

static int flg_mode;
static int flg_break;

static void handlenetoutput(int len)
{
	int i;
	static int nl;

	for (i = 0; i < len; i++) {
		if(flg_mode!=0 && buf[i]=='\033' && (len==1 || buf[i+1]=='\033'))
			flg_break++;
		if(buf[i]=='\n') {
			nl++;
			DEB_TRACE(if(nl==5) msg(stderr, "\n 5!\r\n");)
		}
	}
	if(flg_mode==0) {
#ifdef __CYGWIN__
		buf[len] = 0;
		win_koi8(buf);
#endif
		write(netfd, buf, len);
	}
}


#ifndef __CYGWIN__
static char *files_debug[2] = { "f1", "f2" };
#else
static char cFiles[16384];  /*   */
#endif
static char **files_dim;
static int files_n = -1;

#define FIRST 0
#define NEXT  1
#define CRC   2
#define CLOSE 3
#define PRCNT 5

FILE *iof;

static unsigned long crc_32_tab[] = { /* CRC polynomial 0xedb88320 */
0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

static unsigned long UPDC32(int octet, unsigned long crc)
{
	return crc_32_tab[(crc ^ octet) & 0xff] ^ (crc >> 8);
}

static void send_crc(unsigned long end_pos, char **scrc)
{
	unsigned long oldcrc32 = 0xFFFFFFFF;
	unsigned long pos=0;
	int c;
	static char buf[10];

	fseek(iof, 0l, SEEK_SET);
	while( end_pos > pos++ && (c = getc(iof))!=EOF )
		oldcrc32 = UPDC32(c, oldcrc32);
	oldcrc32 = ~oldcrc32;
	sprintf(buf, "%08lX%c", oldcrc32, (scrc ? 0 : '\n') );
	DEB_TRACE(fprintf(stderr, "client crc = %s\n", buf);)
	if(scrc)
		*scrc = buf;
	 else
		write(netfd, buf, 9);
}

static long szgf;
static long curszgf;
static int get_err;


static void reset_files(void)
{

#ifdef __CYGWIN__
	if(files_n>0) {
		int i;

		for(i=files_n-1; i>=0; i--) {
			if(files_dim[0]!=cFiles)
				free(files_dim[0]);
			files_dim[i-1] = files_dim[i];
		}
		free(files_dim);
	}
#endif
	files_n = -1;
	flg_mode = 0;
	flg_break = 0;
	if(iof!=NULL) {
		fclose(iof);
		iof=NULL;
	}
}

static void choise_put_files(void)
{

	int i;

	DEB_TRACE(fprintf(stderr, "choise_put_files files_n = %d\n", files_n);)

	if(files_n<0) for(;;) {
#ifndef __CYGWIN__
		char buf[16];
		printf("choise_put_files (0,1,2) => ");
		fgets(buf, 16, stdin);
		if(buf[0]!='0' && buf[0]!='1' && buf[0]!='2')
			continue;
		files_n = buf[0]-'0';
		files_dim = files_debug;
#else
		OPENFILENAME ofn;

		memset(&ofn, 0, sizeof(OPENFILENAME));
		ofn.lStructSize = sizeof(OPENFILENAME);
		ofn.Flags = OFN_ALLOWMULTISELECT | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
			OFN_NOLONGNAMES | OFN_PATHMUSTEXIST | OFN_EXPLORER;

		ofn.lpstrInitialDir = upload_dir;
		ofn.lpstrFilter     = "Arj Files\0*.arj\0All files\0*.*\0";
		ofn.nFilterIndex    = 1;
		ofn.lpstrFile       = cFiles;
		ofn.nMaxFile        = sizeof(cFiles);

		files_dim = 0;
		files_n = 0;
		if (GetOpenFileName(&ofn)) {
			files_dim = xmalloc(sizeof(char *));
			if(ofn.nFileExtension!=0) {
				files_dim[files_n++] = cFiles;
			} else {
				int iPosition;
				int ilcurdir;   /*   */

				ilcurdir = iPosition = strlen(cFiles)+1;
				while(1) {
					i = strlen(&cFiles[iPosition]);
					if(i==0)
						break;
					if(files_n)
						files_dim = xrealloc(files_dim, (files_n+1)*sizeof(char *));
					files_dim[files_n] = xmalloc(i+ilcurdir+2);
					sprintf(files_dim[files_n++], "%s\\%s", cFiles, &cFiles[iPosition]);
					iPosition += i+1;
				}
			}
		}
#endif
		break;
	}
	if(files_n==0 || flg_break!=0 || (iof=fopen(files_dim[0], "rb"))==NULL) {
		write(netfd, "-\n", 2);
		reset_files();
	} else {
		char *only_name;

		curszgf = 0;
		fseek(iof, 0l, SEEK_END);
		szgf = ftell(iof);
		fseek(iof, 0l, SEEK_SET);
		only_name = strrchr(files_dim[0],
#ifdef __CYGWIN__
						'\\');
#else
						'/');
#endif
		if(only_name == NULL)
			only_name = files_dim[0];
		 else
			only_name++;
		msg(stdout, "\n%-10ld %-3d%%  %s (%ld)\r", 0l, 0, only_name, szgf);
		fflush(stdout);
		write(netfd, only_name, strlen(only_name));
		write(netfd, "\n", 1);
#ifdef __CYGWIN__
		if(files_dim[0]!=cFiles)
			free(files_dim[0]);
#endif
		for(i=1; i<files_n; i++)
			files_dim[i-1] = files_dim[i];
		files_n--;
	}
}

static void send_blk(void)
{
	char buf[4+1024];
	int n;

	if(flg_break) {
		write(netfd, "-\n", 2);
		reset_files();
		return;
	}
	buf[0] = '+';
	n = fread(buf+3, 1, 1024, iof);
	if(n<=0) {
		if(ferror(iof)) {
			msg(stderr, " I/O %s\n", strerror(errno));
			write(netfd, "-\n", 2);
			}
		 else
			write(netfd, "0", 1);
		return;
		}
	buf[1] = (n>>8);
	buf[2] = (n&0xff);
#ifdef DEBUG
	buf[n+3]=0;
#endif
	DEB_TRACE(fprintf(stderr, "client write = %s\n", buf);)
	write(netfd, buf, n+3);
	curszgf += n;
	n = (int)(100.0*curszgf/(double)szgf);
	msg(stdout, "%-10ld %-3d%%\r", curszgf, n);
}

static int send_a(const char *crcbuf)
{
	int i;
	long offset;
	long offset2;
	char *name;
	char *ep;
	char *scrc;
	char *fullname;

	name = xstrdup(crcbuf);

	for(i=0; crcbuf[i]!=0; i++)
		if(crcbuf[i]==1)
			break;
	if(crcbuf[i]==0)
		err_proto();
	name[i]=0;

	fullname = xmalloc(i+strlen(upload_dir)+1);
	strcpy(fullname, upload_dir);
	strcat(fullname, name);

	iof = fopen(fullname, "r+b");
	if(iof==NULL) {
		msg(stderr, "   %s: %s\n", fullname, strerror(errno));
		write(netfd, "-\n", 2);
		free(name);
		free(fullname);
		return 0;
	}
	free(fullname);
	scrc = &name[++i];
	for(; name[i]!=0; i++)
		if(name[i]=='*')
			break;
	if(name[i]==0)
		err_proto();
	offset = strtol(scrc, &ep, 16);
	if(offset<=0 || *ep != '*') err_proto();
	fseek(iof, 0l, SEEK_END);
	offset2 = ftell(iof);
	if(offset2<offset) {
ne_q:
		fclose(iof);
		iof=NULL;
		msg(stderr, " %s      !\n", name);
		write(netfd, "-\n", 2);
		free(name);
		return 0;
	}
	fseek(iof, 0l, SEEK_SET);
	send_crc(offset, &scrc);
	if(strcmp(scrc, crcbuf+i+1)) goto ne_q;
	curszgf = offset;
	szgf = offset2;
	i = (int)(100.0*curszgf/(double)szgf);
	msg(stdout, "\n%-10ld %-3d%%  %s (%ld)\r", curszgf, i, name, szgf);
	fflush(stdout);
	free(name);
	send_blk();
	return 1;
}


static void send_pos(const char *crcbuf)
{
	char *name = xstrdup(crcbuf);
	char *ep;
	long pos;
	char pb[16];
	int i;
	char *fullname;

	for(i=0; crcbuf[i]!=0; i++)
		if(crcbuf[i]==1)
			break;
	if(crcbuf[i]==0)
		err_proto();
	name[i]=0;

	fullname = xmalloc(i+strlen(download_dir)+1);
	strcpy(fullname, download_dir);
	strcat(fullname, name);

	iof = fopen(fullname, "a+b");
	if(iof==NULL) {
		msg(stderr, "   %s: %s\n", fullname, strerror(errno));
		write(netfd, "-\n", 2);
		free(name);
		free(fullname);
		return;
	}
	free(fullname);
	szgf = strtol(name+i+1, &ep, 16);
	if(szgf<=0 || *ep != 0) err_proto();
	pos = ftell(iof);
	sprintf(pb, "%lx\n", pos);
	write(netfd, pb, strlen(pb));
	get_err=0;
	curszgf = pos;
	i = (int)(100.0*curszgf/(double)szgf);
	msg(stdout, "\n%-10ld %-3d%%  %s (%ld)\r", curszgf, i, name, szgf);
	fflush(stdout);
	free(name);
}

static void read_blk(int mode)
{
	if(mode>=0) {
		if(get_err)
			return;
		putc(mode, iof);
		if(ferror(iof)) {
			msg(stderr, " I/O\n");
			write(netfd, "-\n", 2);
			get_err++;
		} else
		     curszgf++;
		return;
	}
	if(mode == -CLOSE) {
		if(iof) {
			fclose(iof);
			iof=NULL;
		}
		printf("\n");
	}
	if(mode == -CRC) {
		send_crc(curszgf, 0);
	}
	if(mode == -PRCNT) {
		int i = (int)(100.0*curszgf/(double)szgf);
		msg(stdout, "%-10ld %-3d%%\r", curszgf, i);
		fflush(stdout);
		if(flg_break) {
			write(netfd, "-\n", 2);
			flg_break = 0;
		} else
			write(netfd, "\n", 1);
	}
}

static void handlenetinput(int len)
{
	int i;
	unsigned char *start = (unsigned char *)buf;
	static char crcbuf[256+32];
	static int ncrcbuf;
	static int prev_mode;

	for(; len > 0; ) {
		DEB_TRACE(fprintf(stderr, "for(; len=%d\n", len);)
		if(flg_mode==0) {
			DEB_TRACE(fprintf(stderr, "flg_mode==0\n");)
			for (i = 0; i < len; i++) {
				if(start[i]==1) {
					DEB_TRACE(fprintf(stderr, "Detect 1\n");)
					break;
				}
			}
			if(i!=0) {
				strncpy(crcbuf, start, i);
				crcbuf[i]=0;
				msg(stdout, "%s", crcbuf);
				fflush(stdout);
				if(wait_flg==1) {
					int nw = 0;
loop:
					for(; nw<nwait_words; nw++) {
						int nfc = n_found_chars[nw];
						char *combuf = xmalloc(nfc+i+1);
						int lww = strlen(wait_words[nw]);
						char *s = combuf;

						strncpy(combuf, wait_words[nw], nfc);
						strcpy(combuf+nfc, crcbuf);
						while(*s) {
						    int l = strlen(s);

						    if(l>lww)
							l = lww;
						    if(strncmp(s, wait_words[nw], l)==0) {
							free(combuf);
							if(l==lww) {    /* full compared */
								memset(n_found_chars, 0, nwait_words*sizeof(int));
								error_wait_code = nwait_words==1 ? 0 : nw+1;
								wait_flg=2;
								goto test_mode1;
							} else {        /* partial */
								n_found_chars[nw] = l;
								goto loop;
							}
						    }
						    s++;
						}
						free(combuf);
						n_found_chars[nw] = 0;
					}
				}
			}
test_mode1:
			if(i<len) {
				flg_mode='1';
				DEB_TRACE(fprintf(stderr, "set flg_mode to 1\n");)
			}
			i++;
			start += i;
			len -= i;
			DEB_TRACE(fprintf(stderr, "continue # 1, len=%d\n", len);)
			continue;
		}
		if(flg_mode=='1') {
			ncrcbuf = 0;
			flg_mode = '\n';
			prev_mode = *start++;
			if(prev_mode!='A' && prev_mode!='P' && prev_mode!='G')
				err_proto();
			len--;
			DEB_TRACE(fprintf(stderr, "set flg_mode='\\n' prev_mode = %c, len=%d\n", prev_mode, len);)
			continue;
		}
		if(flg_mode=='\n') {
			if(*start != '\n') {
				if(ncrcbuf>=sizeof(crcbuf)) err_proto();
				crcbuf[ncrcbuf++] = *start++;
				len--;
				DEB_TRACE(fprintf(stderr, "continue # 3, len=%d\n", len);)
				continue;
			}
			if(len>1)
				err_proto();
			crcbuf[ncrcbuf]=0;
			flg_mode = prev_mode;
			DEB_TRACE(fprintf(stderr, "continue # 4, flg_mode=%c len=%d\n", flg_mode, len);)
		}
		if(flg_mode=='G') {
			if(ncrcbuf>1) err_proto();
			flg_mode = 'g';
			choise_put_files();
			return;
		}
		if(flg_mode=='g') {
			fflush(stdout);
			i = *start++;
			len--;
			if(i == '-') {
				putchar('\n');
				reset_files();
				iof=NULL;
				continue;
			}
			if(i == '*') {
				ncrcbuf = 0;
				prev_mode = 'C';
				flg_mode = '\n';
				continue;
			}
			if(i != '+' || *start++ != '\n' || len>2) {
				reset_files();
				iof=NULL;
				start--;
				len++;
				continue;
				}
			send_blk();
			DEB_TRACE(fprintf(stderr, "send_blk()\n");)
			return;
		}
		if(flg_mode=='C') {
			if(ncrcbuf==8) {
				send_crc(0xFFFFFFFF, 0);
				fclose(iof);
				iof=NULL;
				flg_mode = 0;
				return;
			} else
				err_proto();
		}
		if(flg_mode=='A') {
			if(send_a(crcbuf)>0)
				flg_mode = 'g';
			 else
				flg_mode = 0;
			return;
		}
		if(flg_mode=='P') {
			send_pos(crcbuf);
			flg_mode = 'p';
			return;
		}
		if(flg_mode=='p') {
			i = *start++;
			len--;
			DEB_TRACE(fprintf(stderr, "flg_mode = %c len=%d i=%c\n", flg_mode, len, i);)
			if(i == '-') {
				read_blk(-CLOSE);
				flg_mode = 0;
				continue;
			}
			if(i == '*') {
				ncrcbuf = 0;
				flg_mode = '\n';
				prev_mode = 'c';
				continue;
			}
			if(i != '+') err_proto();
			flg_mode = 'B';
			DEB_TRACE(fprintf(stderr, "flg_mode = B len = %d continue\n", len);)
			continue;
		}
		if(flg_mode=='B') {
			ncrcbuf = *start++;
			ncrcbuf <<= 8;
			len--;
			flg_mode = '2';
			DEB_TRACE(fprintf(stderr, "flg_mode = 2 len = %d continue\n", len);)
			continue;
		}
		if(flg_mode=='2') {
			ncrcbuf |= *start++;
			len--;
			flg_mode = 'b';
			DEB_TRACE(fprintf(stderr, "flg_mode = b len = %d continue\n", len);)
			continue;
		}
		if(flg_mode=='b') {
			if(ncrcbuf>0) {
				DEB_TRACE(fprintf(stderr, "read_blk(%c)\n", *start);)
				read_blk(*start++);
				ncrcbuf--;
				len--;
				DEB_TRACE(fprintf(stderr, "read_blk done, len=%d\n", len);)
				if(ncrcbuf>0)
					continue;
			}
			/* break */
			read_blk(-PRCNT);
			flg_mode = 'p';
			len--;
			DEB_TRACE(fprintf(stderr, "flg_mode = 'p' write(netfd,'\n',1) done, len=%d\n", len);)
			continue;
		}
		if(flg_mode=='c') {
			if(ncrcbuf>8) err_proto();
			DEB_TRACE(fprintf(stderr, "flg_mode = 'c' start read_blk(-CRC)\n");)
			read_blk(-CRC);
			flg_mode = 'p';
			DEB_TRACE(fprintf(stderr, "end read_blk(-CRC) flg_mode set = 'p'\n");)
			return;
		}
	}
}

static void usage(void) {
	error_msg_and_die("Usage: client [-h host] [-p port] [-d download dir] [-u upload dir] [-s script]\n");
}

static void form_dir(char *arg, char **d)
{
	char *sud;
	int   len = strlen(arg);

	if(len==0)
		return;
	sud = xmalloc(len+2);
	strcpy(sud, arg);
#ifdef __CYGWIN__
	if(sud[len-1]!='\\')
		strcat(sud, "\\");
#else
	if(sud[len-1]!='/')
		strcat(sud, "/");
#endif
	*d = sud;
}


static void set_scr_s(struct SCR_S *set)
{
	int n = n_scr_s;

	scr_s = xrealloc(scr_s, (++n_scr_s)*sizeof(struct SCR_S));
	scr_s[n].cmd = set->cmd;
	scr_s[n].timeout_code = set->timeout_code;
	if(set->word)
		scr_s[n].word = xstrdup(set->word);
	 else
		scr_s[n].word = 0;
}

void parse_src(void)
{
	char buf[256];
	int l;
	int nl=0;
	char *s;
	struct SCR_S set;
	int flg_or = 0;
	char *e;

	while( fgets(buf, sizeof(buf), fscr) != NULL ) {
		nl++;
		l = strlen(buf);
		if(l<=0 || buf[l-1]!='\n')
			error_msg_and_die(" #1  format script file  line %d\n", nl);
		buf[l-1]=0;
#ifdef __CYGWIN__
		if(l>=2 && buf[l-2]=='\r')
			buf[l-2]=0;
#endif
		s = buf;
		if(*s == 0)
			continue;
		while(*s == ' ') s++;
		if(*s == '#')
			continue;
		win_koi8(s);
		for(l=0; rez_words[l]; l++)
			if(strncasecmp(s, rez_words[l], strlen(rez_words[l]))==0)
				break;
		if(rez_words[l]==0) {
err_rw:
			error_msg_and_die("       %d\n", nl);
		}
		s += strlen(rez_words[l]);
		set.cmd = l;
		if(l!=LABEL_SET && l!=SEND_CMD && l!=EXIT_CMD &&
					set.cmd!=ECHO_CMD &&
					set.cmd!=UP_DIR_CMD &&
					set.cmd!=DOWN_DIR_CMD &&
					set.cmd!=HOST_CMD &&
					set.cmd!=PORT_CMD &&
					*s != ' ')
			goto err_rw;
		while(*s == ' ') s++;
		if(set.cmd==UP_DIR_CMD || set.cmd==DOWN_DIR_CMD) {
			if(set.cmd==UP_DIR_CMD) {
				if(up_d_alloc)
					free(upload_dir);
				form_dir(s, &upload_dir);
			} else {
				if(down_d_alloc)
					free(download_dir);
				form_dir(s, &download_dir);
			}
			continue;
		}
		if(set.cmd==HOST_CMD || set.cmd==PORT_CMD) {
			if(set.cmd==HOST_CMD) {
				if(h_alloc)
					free(h);
				h = strdup(s);
			} else {
				if(p_alloc)
					free(porta);
				porta = strdup(s);
			}
			continue;
		}
		if(set.cmd == WAIT_CMD || set.cmd==WAIT_OR_CMD) {
			set.timeout_code = strtol(s, &e, 10);
			if(e == s || *e != ' ' || set.timeout_code < 0)
				error_msg_and_die("  wait    %d\n", nl);
			s = ++e;
			flg_or = set.cmd==WAIT_OR_CMD;
lor:
			while(*s == ' ') s++;
			if(*s == 0 && set.cmd!=SEND_CMD && set.cmd!=ECHO_CMD)
				error_msg_and_die("   %d\n", nl);
			set.word = s;
			set_scr_s(&set);
			continue;
		}
		if(set.cmd == OR_CMD) {
			if(flg_or==0)
				error_msg_and_die("  `or`   %d\n", nl);
			goto lor;
		}
		flg_or = 0;
		if(set.cmd == SEND_CMD || set.cmd == ECHO_CMD) goto lor;
		if(set.cmd == IF_CMD) {
			for(l=0; rez_words_if[l]; l++)
				if(strncasecmp(s, rez_words_if[l], strlen(rez_words_if[l]))==0)
					break;
			if(rez_words_if[l]==0) {
err_rw_if:
				error_msg_and_die("    if  %d\n", nl);
			}
			s += strlen(rez_words_if[l]);
			if(*s != ' ')
				goto err_rw_if;
			if(l==0)        /* error */
				set.timeout_code = -1;
			else if(l==1)        /* success */
				set.timeout_code = 0;
			else {
				set.timeout_code = strtol(s, &e, 10);
				if(e == s || *e != ' ' || set.timeout_code <= 0)
					error_msg_and_die("  if wait_or    %d\n", nl);
				s = ++e;
			}
			while(*s == ' ') s++;
			if(*s == 0)
				error_msg_and_die("     %d\n", nl);
			set.word = s;
			set_scr_s(&set);
			continue;
		}
		if(set.cmd == EXIT_CMD) {
			if(*s == 0)
				set.timeout_code = 0;
			else {
				set.timeout_code = strtol(s, &e, 10);
				if(e == s || *e != 0 || set.timeout_code <= 0)
					error_msg_and_die("  exit  number   %d\n", nl);
			}
			set.word = 0;
			set_scr_s(&set);
			continue;
		}
		if(set.cmd == LABEL_SET) {
			if(*s == 0)
				error_msg_and_die("     %d\n", nl);
			set.word = s;
			set_scr_s(&set);
			continue;
		}
	}
	fclose(fscr);
	for(l=0; l<n_scr_s; l++) {
		if(scr_s[l].cmd == IF_CMD) {
			for(nl=0; nl<n_scr_s; nl++)
				if(scr_s[nl].cmd == LABEL_SET &&
						strcmp(scr_s[nl].word, scr_s[l].word)==0)
					break;
			if(nl==n_scr_s)
				error_msg_and_die(" %s  \n", scr_s[l].word);
		}
	}
}

static void next_cur_scr(void)
{
	if(cur_scr) {
		cur_scr++;
		if( cur_scr >= (scr_s+n_scr_s) )
			cur_scr = 0;
	}
}

static void setup_wait_words(void)
{
	struct SCR_S *sav_cur_scr = cur_scr;

	for(nwait_words = 1;
		next_cur_scr(), (cur_scr!=0 && cur_scr->cmd == OR_CMD);
			nwait_words++)
				;
	cur_scr = sav_cur_scr;
	wait_words = xrealloc(wait_words, nwait_words*sizeof(char *));
	n_found_chars = xrealloc(n_found_chars, nwait_words*sizeof(int));

	nwait_words = 0;
	do {
		wait_words[nwait_words] = cur_scr->word;
		n_found_chars[nwait_words] = 0;
		next_cur_scr();
		nwait_words++;
	} while(cur_scr!=0 && cur_scr->cmd == OR_CMD);

	cur_scr = sav_cur_scr;
}

int main(int argc, char** argv)
{
	struct in_addr host;
	int port;
	fd_set readfds;
	int maxfd;
	struct hostent * he;
	int len;
	int option;

	time_t tt;

	while ((option = getopt(argc, argv, "h:p:d:u:s:")) > 0) {
		switch (option) {
		case 'h':
			h = xstrdup(optarg);
			h_alloc = 1;
			break;
		case 'p':
			porta = xstrdup(optarg);
			p_alloc = 1;
			break;
		case 'd':
			form_dir(optarg, &download_dir);
			down_d_alloc=1;
			break;
		case 'u':
			form_dir(optarg, &upload_dir);
			up_d_alloc=1;
			break;
		case 's':
			strscript = xstrdup(optarg);
			s_allocated = 1;
			break;
		default:
			usage();

		}
	}
	if (optind!=argc)
	    usage();

	if(s_allocated) {
		fscr = fopen(strscript, "r");
		if(fscr==NULL)
			error_msg_and_die("     %s\n", strscript);
		parse_src();
		if(n_scr_s == 0) {
			if(s_allocated)
				free(strscript);
		} else {
			cur_scr = scr_s;
		}
	}

	if (tcgetattr(0, &termios_def) < 0)
		error_msg_and_die("tcgetattr: %s\n", strerror(errno));

	termios_raw = termios_def;

	termios_raw.c_cc[VMIN] = 1;
	termios_raw.c_cc[VTIME] = 0;
	termios_raw.c_cc[VINTR] = _POSIX_VDISABLE;     /* Turn off CTRL-C, so we can trap it */
	termios_raw.c_lflag &= ~ICANON;        /* unbuffered input */
	termios_raw.c_lflag &= ~(ECHO | ECHOCTL | ECHONL);     /* Turn off echoing */

	port = atoi(porta);
	if ((unsigned)(port - 1 ) > 65534)
		error_msg_and_die("%s:   \n", argv[2]);

	if ((he = gethostbyname(h)) == NULL)
		error_msg_and_die("%s:  \n", h);

	memcpy(&host, he->h_addr, sizeof(struct in_addr));

	netfd = remote_connect(host, port);

	tcsetattr(0, TCSADRAIN, &termios_raw);
	term_inited++;

	FD_ZERO(&readfds);
	FD_SET(0, &readfds);
	FD_SET(netfd, &readfds);
	maxfd = netfd + 1;

	while (1) {
		fd_set rfds = readfds;
		struct timeval tv;
		struct timeval *ptv = 0;

		if(cur_scr!=0 && flg_mode==0) {
			if(cur_scr->cmd==SEND_CMD) {
				if(*(cur_scr->word))
					write(netfd, cur_scr->word, strlen(cur_scr->word));
				write(netfd, "\n", 1);
				next_cur_scr();
				continue;
			}
			if(cur_scr->cmd==LABEL_SET) {
				next_cur_scr();
				continue;
			}
			if(cur_scr->cmd==ECHO_CMD) {
				if(*(cur_scr->word))
					msg(stdout, "%s", cur_scr->word);
				printf("\n");
				next_cur_scr();
				continue;
			}
			if(cur_scr->cmd==EXIT_CMD) {
				if(cur_scr->timeout_code > 0)
					doexit(cur_scr->timeout_code);
				cur_scr = 0;
				continue;
			}
			if(cur_scr->cmd==WAIT_CMD || cur_scr->cmd==WAIT_OR_CMD) {
				if(wait_flg==0) {
					wait_flg = 1;
					if(cur_scr->timeout_code>0) {
						time(&tt);
						ptv = &tv;
						tv.tv_sec = cur_scr->timeout_code;
						tv.tv_usec = 0;
					}
					setup_wait_words();
				} else if(wait_flg==1) { /* continue finding */
					if(cur_scr->timeout_code>0) {
						time_t tt2;

						time(&tt2);
						ptv = &tv;
						tv.tv_usec = 0;
						tv.tv_sec = cur_scr->timeout_code-(tt2-tt);
						if(tv.tv_sec < 0) {
							ptv = 0;
							wait_flg = -1;
							error_wait_code = -1;
							continue;
						}
					}
				} else  { /* timeout or found */
					do
						next_cur_scr();
					while(cur_scr!=0 && cur_scr->cmd==OR_CMD);
					wait_flg = 0;
					continue;
				}
			}
			if(cur_scr->cmd==IF_CMD) {
				if(cur_scr->timeout_code!=error_wait_code) {
					next_cur_scr();
					continue;
				}
				for(len = 0; len < n_scr_s; len++)
					if(scr_s[len].cmd == LABEL_SET &&
						!strcmp(scr_s[len].word, cur_scr->word))
							break;
				cur_scr = &scr_s[len];
				continue;
			}
		}
		switch (select(maxfd, &rfds, NULL, NULL, ptv)) {
		case 0:
			/* timeout */
			if(wait_flg) {
				error_wait_code = -1;
				wait_flg = -1;
			}
		case -1:
			/* error, ignore, bay go to loop */
			sleep(1);
			break;
		default:

			if (FD_ISSET(0, &rfds)) {
				len = read(0, buf, DATABUFSIZE
#ifdef __CYGWIN__
					-1
#endif
					);

				if (len <= 0)
					doexit(0);
				cur_scr = 0;
				handlenetoutput(len);
			}

			if (FD_ISSET(netfd, &rfds)) {
				len = read(netfd, buf, DATABUFSIZE);

				if (len <= 0)
					error_msg_and_die("Connection closed by foreign host.\r\n");

				handlenetinput(len);
			}
		}
	}
}
